"""GEM is the intermediate language of TSFC for describing
tensor-valued mathematical expressions and tensor operations.
It is similar to Einstein's notation.

Its design was heavily inspired by UFL, with some major differences:
 - GEM has got nothing FEM-specific.
 - In UFL free indices are just unrolled shape, thus UFL is very
   restrictive about operations on expressions with different sets of
   free indices. GEM is much more relaxed about free indices.

Similarly to UFL, all GEM nodes have 'shape' and 'free_indices'
attributes / properties. Unlike UFL, however, index extents live on
the Index objects in GEM, not on all the nodes that have those free
indices.
"""

from abc import ABCMeta
from itertools import chain
from functools import reduce
from operator import attrgetter
from numbers import Integral, Number

import numpy
from numpy import asarray

from gem.node import Node as NodeBase, traversal

from FIAT.orientation_utils import Orientation as FIATOrientation


__all__ = ['Node', 'Identity', 'Literal', 'Zero', 'Failure',
           'Variable', 'Sum', 'Product', 'Division', 'FloorDiv', 'Remainder', 'Power',
           'MathFunction', 'MinValue', 'MaxValue', 'Comparison',
           'LogicalNot', 'LogicalAnd', 'LogicalOr', 'Conditional',
           'Index', 'VariableIndex', 'Indexed', 'ComponentTensor',
           'IndexSum', 'ListTensor', 'Concatenate', 'Delta', 'OrientationVariableIndex',
           'index_sum', 'partial_indexed', 'reshape', 'view',
           'indices', 'as_gem', 'FlexiblyIndexed',
           'Inverse', 'Solve', 'extract_type', 'uint_type']


uint_type = numpy.dtype(numpy.uintc)


class NodeMeta(type):
    """Metaclass of GEM nodes.

    When a GEM node is constructed, this metaclass automatically
    collects its free indices if 'free_indices' has not been set yet.
    """

    def __call__(self, *args, **kwargs):
        # Create and initialise object
        obj = super(NodeMeta, self).__call__(*args, **kwargs)

        # Set free_indices if not set already
        if not hasattr(obj, 'free_indices'):
            obj.free_indices = unique(chain(*[c.free_indices
                                              for c in obj.children]))
        # Set dtype if not set already.
        if not hasattr(obj, 'dtype'):
            obj.dtype = obj.inherit_dtype_from_children(obj.children)

        return obj


class Node(NodeBase, metaclass=NodeMeta):
    """Abstract GEM node class."""

    __slots__ = ('free_indices', 'dtype')

    def is_equal(self, other):
        """Common subexpression eliminating equality predicate.

        When two (sub)expressions are equal, the children of one
        object are reassigned to the children of the other, so some
        duplicated subexpressions are eliminated.
        """
        result = NodeBase.is_equal(self, other)
        if result:
            self.children = other.children
        return result

    def __getitem__(self, indices):
        try:
            indices = tuple(indices)
        except TypeError:
            indices = (indices, )
        return Indexed(self, indices)

    def __add__(self, other):
        return componentwise(Sum, self, as_gem(other))

    def __radd__(self, other):
        return as_gem(other).__add__(self)

    def __sub__(self, other):
        return componentwise(
            Sum, self,
            componentwise(Product, Literal(-1), as_gem(other)))

    def __rsub__(self, other):
        return as_gem(other).__sub__(self)

    def __mul__(self, other):
        return componentwise(Product, self, as_gem(other))

    def __rmul__(self, other):
        return as_gem(other).__mul__(self)

    def __matmul__(self, other):
        other = as_gem(other)
        if not self.shape and not other.shape:
            return Product(self, other)
        elif not (self.shape and other.shape):
            raise ValueError("Both objects must have shape for matmul")
        elif self.shape[-1] != other.shape[0]:
            raise ValueError(f"Mismatching shapes {self.shape} and {other.shape} in matmul")
        *i, k = indices(len(self.shape))
        _, *j = indices(len(other.shape))
        expr = Product(Indexed(self, tuple(i) + (k, )),
                       Indexed(other, (k, ) + tuple(j)))
        return ComponentTensor(IndexSum(expr, (k, )), tuple(i) + tuple(j))

    def __rmatmul__(self, other):
        return as_gem(other).__matmul__(self)

    @property
    def T(self):
        i = indices(len(self.shape))
        return ComponentTensor(Indexed(self, i), tuple(reversed(i)))

    def __truediv__(self, other):
        other = as_gem(other)
        if other.shape:
            raise ValueError("Denominator must be scalar")
        return componentwise(Division, self, other)

    def __rtruediv__(self, other):
        return as_gem(other).__truediv__(self)

    def __floordiv__(self, other):
        other = as_gem_uint(other)
        if other.shape:
            raise ValueError("Denominator must be scalar")
        return componentwise(FloorDiv, self, other)

    def __rfloordiv__(self, other):
        return as_gem_uint(other).__floordiv__(self)

    def __mod__(self, other):
        other = as_gem_uint(other)
        if other.shape:
            raise ValueError("Denominator must be scalar")
        return componentwise(Remainder, self, other)

    def __rmod__(self, other):
        return as_gem_uint(other).__mod__(self)

    @staticmethod
    def inherit_dtype_from_children(children):
        if any(c.dtype is None for c in children):
            # Set dtype = None will let _assign_dtype()
            # assign the default dtype for this node later.
            return
        else:
            return numpy.result_type(*(c.dtype for c in children))


class Terminal(Node):
    """Abstract class for terminal GEM nodes."""

    __slots__ = ('_dtype',)

    children = ()

    is_equal = NodeBase.is_equal

    @property
    def dtype(self):
        """Data type of the node.

        We only need to set dtype (or _dtype) on terminal nodes, and
        other nodes inherit dtype from their children.

        Currently dtype is significant only for nodes under index DAGs
        (DAGs underneath `VariableIndex`s representing indices), and
        `VariableIndex` checks if the dtype of the node that it wraps is
        of uint_type. _assign_dtype() will then assign uint_type to those nodes.

        dtype can be `None` otherwise, and _assign_dtype() will assign
        the default dtype to those nodes.

        """
        if hasattr(self, '_dtype'):
            return self._dtype
        else:
            raise AttributeError(f"Must set _dtype on terminal node, {type(self)}")


class Scalar(Node):
    """Abstract class for scalar-valued GEM nodes."""

    __slots__ = ()

    shape = ()


class Failure(Terminal):
    """Abstract class for failure GEM nodes."""

    __slots__ = ('shape', 'exception')
    __front__ = ('shape', 'exception')

    def __init__(self, shape, exception):
        self.shape = shape
        self.exception = exception
        self._dtype = None


class Constant(Terminal):
    """Abstract base class for constant types.

    Convention:
     - array: numpy array of values
     - value: float or complex value (scalars only)
    """
    pass


class Zero(Constant):
    """Symbolic zero tensor"""

    __slots__ = ('shape',)
    __front__ = ('shape',)
    __back__ = ('dtype',)

    def __init__(self, shape=(), dtype=None):
        self.shape = shape
        self._dtype = dtype

    @property
    def value(self):
        assert not self.shape
        return numpy.array(0, dtype=self.dtype or float).item()


class Identity(Constant):
    """Identity matrix"""

    __slots__ = ('dim',)
    __front__ = ('dim',)
    __back__ = ('dtype',)

    def __init__(self, dim, dtype=None):
        self.dim = dim
        self._dtype = dtype

    @property
    def shape(self):
        return (self.dim, self.dim)

    @property
    def array(self):
        return numpy.eye(self.dim, dtype=self.dtype)


class Literal(Constant):
    """Tensor-valued constant"""

    __slots__ = ('array',)
    __front__ = ('array',)
    __back__ = ('dtype',)

    def __new__(cls, array, dtype=None):
        array = asarray(array)
        return super(Literal, cls).__new__(cls)

    def __init__(self, array, dtype=None):
        array = asarray(array)
        if dtype is None:
            # Assume float or complex.
            try:
                self.array = array.astype(float, casting="safe")
            except TypeError:
                self.array = array.astype(complex)
        else:
            # Can be int, etc.
            self.array = array.astype(dtype)
        self._dtype = self.array.dtype

    def is_equal(self, other):
        if type(self) is not type(other):
            return False
        if self.shape != other.shape:
            return False
        return tuple(self.array.flat) == tuple(other.array.flat)

    def get_hash(self):
        return hash((type(self), self.shape, tuple(self.array.flat)))

    @property
    def value(self):
        assert self.shape == ()
        return self.array.dtype.type(self.array)

    @property
    def shape(self):
        return self.array.shape


class Variable(Terminal):
    """Symbolic variable tensor"""

    __slots__ = ('name', 'shape')
    __front__ = ('name', 'shape')
    __back__ = ('dtype',)

    def __init__(self, name, shape, dtype=None):
        self.name = name
        self.shape = shape
        self._dtype = dtype


class Sum(Scalar):
    __slots__ = ('children',)

    def __new__(cls, *args):
        try:
            a, b = args
        except ValueError:
            # Handle more than two arguments
            return reduce(Sum, args)
        assert not a.shape
        assert not b.shape

        # Constant folding
        if isinstance(a, Zero):
            return b
        elif isinstance(b, Zero):
            return a

        if isinstance(a, Constant) and isinstance(b, Constant):
            return Literal(a.value + b.value, dtype=Node.inherit_dtype_from_children([a, b]))

        self = super(Sum, cls).__new__(cls)
        self.children = a, b
        return self


class Product(Scalar):
    __slots__ = ('children',)

    def __new__(cls, *args):
        try:
            a, b = args
        except ValueError:
            # Handle more than two arguments
            return one if len(args) == 0 else reduce(Product, args)
        assert not a.shape
        assert not b.shape

        # Constant folding
        if isinstance(a, Zero) or isinstance(b, Zero):
            return Zero()

        if a == one:
            return b
        if b == one:
            return a

        if isinstance(a, Constant) and isinstance(b, Constant):
            return Literal(a.value * b.value, dtype=Node.inherit_dtype_from_children([a, b]))

        self = super(Product, cls).__new__(cls)
        self.children = a, b
        return self


class Division(Scalar):
    __slots__ = ('children',)

    def __new__(cls, a, b):
        assert not a.shape
        assert not b.shape

        # Constant folding
        if isinstance(b, Zero):
            raise ValueError("division by zero")
        if isinstance(a, Zero):
            return Zero()

        if b == one:
            return a

        if isinstance(a, Constant) and isinstance(b, Constant):
            return Literal(a.value / b.value, dtype=Node.inherit_dtype_from_children([a, b]))

        self = super(Division, cls).__new__(cls)
        self.children = a, b
        return self


class FloorDiv(Scalar):
    __slots__ = ('children',)

    def __new__(cls, a, b):
        assert not a.shape
        assert not b.shape
        dtype = Node.inherit_dtype_from_children([a, b])
        if dtype != uint_type:
            raise ValueError(f"dtype ({dtype}) != unit_type ({uint_type})")
        # Constant folding
        if isinstance(b, Zero):
            raise ValueError("division by zero")
        if isinstance(a, Zero):
            return Zero(dtype=dtype)
        if isinstance(b, Constant) and b.value == 1:
            return a
        if isinstance(a, Constant) and isinstance(b, Constant):
            return Literal(a.value // b.value, dtype=dtype)
        self = super(FloorDiv, cls).__new__(cls)
        self.children = a, b
        return self


class Remainder(Scalar):
    __slots__ = ('children',)

    def __new__(cls, a, b):
        assert not a.shape
        assert not b.shape
        dtype = Node.inherit_dtype_from_children([a, b])
        if dtype != uint_type:
            raise ValueError(f"dtype ({dtype}) != uint_type ({uint_type})")
        # Constant folding
        if isinstance(b, Zero):
            raise ValueError("division by zero")
        if isinstance(a, Zero):
            return Zero(dtype=dtype)
        if isinstance(b, Constant) and b.value == 1:
            return Zero(dtype=dtype)
        if isinstance(a, Constant) and isinstance(b, Constant):
            return Literal(a.value % b.value, dtype=dtype)
        self = super(Remainder, cls).__new__(cls)
        self.children = a, b
        return self


class Power(Scalar):
    __slots__ = ('children',)

    def __new__(cls, base, exponent):
        assert not base.shape
        assert not exponent.shape
        dtype = Node.inherit_dtype_from_children([base, exponent])

        # Constant folding
        if isinstance(base, Zero):
            if isinstance(exponent, Zero):
                raise ValueError("cannot solve 0^0")
            return Zero(dtype=dtype)
        elif isinstance(exponent, Zero):
            return Literal(1, dtype=dtype)
        elif isinstance(base, Constant) and isinstance(exponent, Constant):
            return Literal(base.value ** exponent.value, dtype=dtype)

        self = super(Power, cls).__new__(cls)
        self.children = base, exponent
        return self


class MathFunction(Scalar):
    __slots__ = ('name', 'children')
    __front__ = ('name',)

    def __new__(cls, name, *args):
        assert isinstance(name, str)
        assert all(arg.shape == () for arg in args)

        if name in {'conj', 'real', 'imag'}:
            arg, = args
            if isinstance(arg, Zero):
                return arg

        self = super(MathFunction, cls).__new__(cls)
        self.name = name
        self.children = args
        return self


class MinValue(Scalar):
    __slots__ = ('children',)

    def __init__(self, a, b):
        assert not a.shape
        assert not b.shape

        self.children = a, b


class MaxValue(Scalar):
    __slots__ = ('children',)

    def __init__(self, a, b):
        assert not a.shape
        assert not b.shape

        self.children = a, b


class Comparison(Scalar):
    __slots__ = ('operator', 'children')
    __front__ = ('operator',)

    def __init__(self, op, a, b):
        assert not a.shape
        assert not b.shape

        if op not in [">", ">=", "==", "!=", "<", "<="]:
            raise ValueError("invalid operator")

        self.operator = op
        self.children = a, b
        self.dtype = None  # Do not inherit dtype from children.


class LogicalNot(Scalar):
    __slots__ = ('children',)

    def __init__(self, expression):
        assert not expression.shape

        self.children = expression,


class LogicalAnd(Scalar):
    __slots__ = ('children',)

    def __init__(self, a, b):
        assert not a.shape
        assert not b.shape

        self.children = a, b


class LogicalOr(Scalar):
    __slots__ = ('children',)

    def __init__(self, a, b):
        assert not a.shape
        assert not b.shape

        self.children = a, b


class Conditional(Node):
    __slots__ = ('children', 'shape')

    def __new__(cls, condition, then, else_):
        assert not condition.shape
        assert then.shape == else_.shape == ()

        # If both branches are the same, just return one of them.  In
        # particular, this will help constant-fold zeros.
        if then == else_:
            return then

        self = super(Conditional, cls).__new__(cls)
        self.children = condition, then, else_
        self.shape = then.shape
        self.dtype = Node.inherit_dtype_from_children([then, else_])
        return self


class IndexBase(metaclass=ABCMeta):
    """Abstract base class for indices."""
    pass


IndexBase.register(int)


class Index(IndexBase):
    """Free index"""

    # Not true object count, just for naming purposes
    _count = 0

    __slots__ = ('name', 'extent', 'count')

    def __init__(self, name=None, extent=None):
        self.name = name
        Index._count += 1
        self.count = Index._count
        self.extent = extent

    def set_extent(self, value):
        # Set extent, check for consistency
        if self.extent is None:
            self.extent = value
        elif self.extent != value:
            raise ValueError("Inconsistent index extents!")

    def __str__(self):
        if self.name is None:
            return "i_%d" % self.count
        return self.name

    def __repr__(self):
        if self.name is None:
            return "Index(%r)" % self.count
        return "Index(%r)" % self.name

    def __lt__(self, other):
        # Allow sorting of free indices in Python 3
        return id(self) < id(other)

    def __getstate__(self):
        return self.name, self.extent, self.count

    def __setstate__(self, state):
        self.name, self.extent, self.count = state


class VariableIndex(IndexBase):
    """An index that is constant during a single execution of the
    kernel, but whose value is not known at compile time."""

    __slots__ = ('expression',)

    def __init__(self, expression):
        assert isinstance(expression, Node)
        assert not expression.shape
        if expression.dtype != uint_type:
            raise ValueError(f"expression.dtype ({expression.dtype}) != uint_type ({uint_type})")
        self.expression = expression

    def __eq__(self, other):
        if self is other:
            return True
        if type(self) is not type(other):
            return False
        return self.expression == other.expression

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash((type(self), self.expression))

    def __str__(self):
        return str(self.expression)

    def __repr__(self):
        return "%r(%r)" % (type(self), self.expression,)

    def __reduce__(self):
        return type(self), (self.expression,)


class Indexed(Scalar):
    __slots__ = ('children', 'multiindex', 'indirect_children')
    __back__ = ('multiindex',)

    def __new__(cls, aggregate, multiindex):
        # Accept numpy or any integer, but cast to int.
        multiindex = tuple(int(i) if isinstance(i, Integral) else i
                           for i in multiindex)

        # Set index extents from shape
        assert len(aggregate.shape) == len(multiindex)
        for index, extent in zip(multiindex, aggregate.shape):
            assert isinstance(index, IndexBase)
            if isinstance(index, Index):
                index.set_extent(extent)
            elif isinstance(index, int) and not (0 <= index < extent):
                raise IndexError("Invalid literal index")

        # Empty multiindex
        if not multiindex:
            return aggregate

        # Zero folding
        if isinstance(aggregate, Zero):
            return Zero(dtype=aggregate.dtype)

        # All indices fixed
        if all(isinstance(i, int) for i in multiindex):
            if isinstance(aggregate, Constant):
                return Literal(aggregate.array[multiindex], dtype=aggregate.dtype)
            elif isinstance(aggregate, ListTensor):
                return aggregate.array[multiindex]

        self = super(Indexed, cls).__new__(cls)
        self.children = (aggregate,)
        self.multiindex = multiindex
        self.indirect_children = tuple(i.expression for i in self.multiindex if isinstance(i, VariableIndex))

        new_indices = []
        for i in multiindex:
            if isinstance(i, Index):
                new_indices.append(i)
            elif isinstance(i, VariableIndex):
                new_indices.extend(i.expression.free_indices)
        self.free_indices = unique(aggregate.free_indices + tuple(new_indices))

        return self

    def index_ordering(self):
        """Running indices in the order of indexing in this node."""
        free_indices = []
        for i in self.multiindex:
            if isinstance(i, Index):
                free_indices.append(i)
            elif isinstance(i, VariableIndex):
                free_indices.extend(i.expression.free_indices)
        return tuple(free_indices)


class FlexiblyIndexed(Scalar):
    """Flexible indexing of :py:class:`Variable`s to implement views and
    reshapes (splitting dimensions only)."""

    __slots__ = ('children', 'dim2idxs', 'indirect_children')
    __back__ = ('dim2idxs',)

    def __init__(self, variable, dim2idxs):
        """Construct a flexibly indexed node.

        Parameters
        ----------
        variable : Node
            `Node` that has a shape.
        dim2idxs : tuple
            Tuple of (offset, ((index, stride), (...), ...)) mapping indices,
            where offset is {Node, int}, index is {Index, VariableIndex, int}, and
            stride is {Node, int}.

        For example, if ``variable`` is rank two, and ``dim2idxs`` is

            ((1, ((i, 12), (j, 4), (k, 1))), (0, ()))

        then this corresponds to the indexing:

            variable[1 + i*12 + j*4 + k][0]

        """
        assert variable.shape
        assert len(variable.shape) == len(dim2idxs)
        dim2idxs_ = []
        free_indices = []
        for dim, (offset, idxs) in zip(variable.shape, dim2idxs):
            offset_ = offset
            idxs_ = []
            last = 0
            if isinstance(offset, Node):
                free_indices.extend(offset.free_indices)
            for index, stride in idxs:
                if isinstance(index, Index):
                    assert index.extent is not None
                    free_indices.append(index)
                    idxs_.append((index, stride))
                    last += (index.extent - 1) * stride
                elif isinstance(index, VariableIndex):
                    base_indices = index.expression.free_indices
                    assert all(base_index.extent is not None for base_index in base_indices)
                    free_indices.extend(base_indices)
                    idxs_.append((index, stride))
                    # last += (unknown_extent - 1) * stride
                elif isinstance(index, int):
                    # TODO: Attach dtype to each Node.
                    # Here, we should simply be able to do:
                    # >>> offset_ += index * stride
                    # but "+" and "*" are not currently correctly overloaded
                    # for indices (integers); they assume floats.
                    if not isinstance(offset, Integral):
                        raise NotImplementedError(f"Found non-Integral offset : {offset}")
                    if isinstance(stride, Constant):
                        offset_ += index * stride.value
                    else:
                        offset_ += index * stride
                else:
                    raise ValueError("Unexpected index type for flexible indexing")
                if isinstance(stride, Node):
                    free_indices.extend(stride.free_indices)
            if dim is not None and isinstance(offset_ + last, Integral) and offset_ + last >= dim:
                raise ValueError("Offset {0} and indices {1} exceed dimension {2}".format(offset, idxs, dim))
            dim2idxs_.append((offset_, tuple(idxs_)))
        self.children = (variable,)
        self.dim2idxs = tuple(dim2idxs_)
        self.free_indices = unique(free_indices)
        indirect_children = []
        for offset, idxs in self.dim2idxs:
            if isinstance(offset, Node):
                indirect_children.append(offset)
            for idx, stride in idxs:
                if isinstance(idx, VariableIndex):
                    indirect_children.append(idx.expression)
                if isinstance(stride, Node):
                    indirect_children.append(stride)
        self.indirect_children = tuple(indirect_children)

    def index_ordering(self):
        """Running indices in the order of indexing in this node."""
        free_indices = []
        for offset, idxs in self.dim2idxs:
            if isinstance(offset, Node):
                free_indices.extend(offset.free_indices)
            for index, stride in idxs:
                if isinstance(index, Index):
                    free_indices.append(index)
                elif isinstance(index, VariableIndex):
                    free_indices.extend(index.expression.free_indices)
                if isinstance(stride, Node):
                    free_indices.extend(stride.free_indices)
        return tuple(free_indices)


class ComponentTensor(Node):
    __slots__ = ('children', 'multiindex', 'shape')
    __back__ = ('multiindex',)

    def __new__(cls, expression, multiindex):
        assert not expression.shape

        # Empty multiindex
        if not multiindex:
            return expression

        # Collect shape
        shape = tuple(index.extent for index in multiindex)
        assert all(s >= 0 for s in shape)

        # Zero folding
        if isinstance(expression, Zero):
            return Zero(shape, dtype=expression.dtype)

        self = super(ComponentTensor, cls).__new__(cls)
        self.children = (expression,)
        self.multiindex = multiindex
        self.shape = shape

        # Collect free indices
        assert set(multiindex) <= set(expression.free_indices)
        self.free_indices = unique(set(expression.free_indices) - set(multiindex))

        return self


class IndexSum(Scalar):
    __slots__ = ('children', 'multiindex')
    __back__ = ('multiindex',)

    def __new__(cls, summand, multiindex):
        # Sum zeros
        assert not summand.shape
        if isinstance(summand, Zero):
            return summand

        # Unroll singleton sums
        unroll = tuple(index for index in multiindex if index.extent <= 1)
        if unroll:
            assert numpy.prod([index.extent for index in unroll]) == 1
            summand = Indexed(ComponentTensor(summand, unroll),
                              (0,) * len(unroll))
            multiindex = tuple(index for index in multiindex
                               if index not in unroll)

        # No indices case
        multiindex = tuple(multiindex)
        if not multiindex:
            return summand

        self = super(IndexSum, cls).__new__(cls)
        self.children = (summand,)
        self.multiindex = multiindex

        # Collect shape and free indices
        assert set(multiindex) <= set(summand.free_indices)
        self.free_indices = unique(set(summand.free_indices) - set(multiindex))

        return self


class ListTensor(Node):
    __slots__ = ('array',)

    def __new__(cls, array):
        array = asarray(array)
        assert numpy.prod(array.shape)
        dtype = Node.inherit_dtype_from_children(tuple(array.flat))

        # Handle children with shape
        child_shape = array.flat[0].shape
        assert all(elem.shape == child_shape for elem in array.flat)

        if child_shape:
            # Destroy structure
            direct_array = numpy.empty(array.shape + child_shape, dtype=object)
            for alpha in numpy.ndindex(array.shape):
                for beta in numpy.ndindex(child_shape):
                    direct_array[alpha + beta] = Indexed(array[alpha], beta)
            array = direct_array

        # Constant folding
        if all(isinstance(elem, Constant) for elem in array.flat):
            return Literal(numpy.vectorize(attrgetter('value'))(array), dtype=dtype)

        self = super(ListTensor, cls).__new__(cls)
        self.array = array
        return self

    @property
    def children(self):
        return tuple(self.array.flat)

    @property
    def shape(self):
        return self.array.shape

    def __reduce__(self):
        return type(self), (self.array,)

    def reconstruct(self, *args):
        return ListTensor(asarray(args).reshape(self.array.shape))

    def __repr__(self):
        return "ListTensor(%r)" % self.array.tolist()

    def is_equal(self, other):
        """Common subexpression eliminating equality predicate."""
        if type(self) is not type(other):
            return False
        if (self.array == other.array).all():
            self.array = other.array
            return True
        return False

    def get_hash(self):
        return hash((type(self), self.shape, self.children))


class Concatenate(Node):
    """Flattens and concatenates GEM expressions by shape.

    Similar to what UFL MixedElement does to value shape.  For
    example, if children have shapes (2, 2), (), and (3,) then the
    concatenated expression has shape (8,).
    """
    __slots__ = ('children',)

    def __new__(cls, *children):
        dtype = Node.inherit_dtype_from_children(children)
        if all(isinstance(child, Zero) for child in children):
            size = int(sum(numpy.prod(child.shape, dtype=int) for child in children))
            return Zero((size,), dtype=dtype)

        self = super(Concatenate, cls).__new__(cls)
        self.children = children
        return self

    @property
    def shape(self):
        return (int(sum(numpy.prod(child.shape, dtype=int) for child in self.children)),)


class Delta(Scalar, Terminal):
    __slots__ = ('i', 'j')
    __front__ = ('i', 'j')
    __back__ = ('dtype',)

    def __new__(cls, i, j, dtype=None):
        if isinstance(i, tuple) and isinstance(j, tuple):
            # Handle multiindices
            return Product(*map(Delta, i, j))
        assert isinstance(i, IndexBase)
        assert isinstance(j, IndexBase)

        # \delta_{i,i} = 1
        if i == j:
            return one

        # Fixed indices
        if isinstance(i, Integral) and isinstance(j, Integral):
            return one if i == j else Zero()

        if isinstance(i, Integral):
            return Indexed(Literal(numpy.eye(j.extent)[i]), (j,))

        if isinstance(j, Integral):
            return Indexed(Literal(numpy.eye(i.extent)[j]), (i,))

        self = super(Delta, cls).__new__(cls)
        self.i = i
        self.j = j
        # Set up free indices
        free_indices = []
        for index in (i, j):
            if isinstance(index, Index):
                free_indices.append(index)
            elif isinstance(index, VariableIndex):
                raise NotImplementedError("Can not make Delta with VariableIndex")
        self.free_indices = tuple(unique(free_indices))
        self._dtype = dtype
        return self


class Inverse(Node):
    """The inverse of a square matrix."""
    __slots__ = ('children', 'shape')

    def __new__(cls, tensor):
        assert len(tensor.shape) == 2
        assert tensor.shape[0] == tensor.shape[1]

        # Invert 1x1 matrix
        if tensor.shape == (1, 1):
            multiindex = (Index(), Index())
            return ComponentTensor(Division(one, Indexed(tensor, multiindex)), multiindex)

        self = super(Inverse, cls).__new__(cls)
        self.children = (tensor,)
        self.shape = tensor.shape

        return self


class Solve(Node):
    """Solution of a square matrix equation with (potentially) multiple right hand sides.

    Represents the X obtained by solving AX = B.
    """
    __slots__ = ('children', 'shape')

    def __init__(self, A, B):
        # Shape requirements
        assert B.shape
        assert len(A.shape) == 2
        assert A.shape[0] == A.shape[1]
        assert A.shape[0] == B.shape[0]

        self.children = (A, B)
        self.shape = A.shape[1:] + B.shape[1:]


class OrientationVariableIndex(VariableIndex, FIATOrientation):
    """VariableIndex representing a fiat orientation.

    Notes
    -----
    In the current implementation, we need to extract
    `VariableIndex.expression` as index arithmetic
    is not supported (indices are not `Node`).

    """

    def __floordiv__(self, other):
        other = other.expression if isinstance(other, VariableIndex) else as_gem_uint(other)
        return type(self)(FloorDiv(self.expression, other))

    def __rfloordiv__(self, other):
        other = other.expression if isinstance(other, VariableIndex) else as_gem_uint(other)
        return type(self)(FloorDiv(other, self.expression))

    def __mod__(self, other):
        other = other.expression if isinstance(other, VariableIndex) else as_gem_uint(other)
        return type(self)(Remainder(self.expression, other))

    def __rmod__(self, other):
        other = other.expression if isinstance(other, VariableIndex) else as_gem_uint(other)
        return type(self)(Remainder(other, self.expression))


def unique(indices):
    """Sorts free indices and eliminates duplicates.

    :arg indices: iterable of indices
    :returns: sorted tuple of unique free indices
    """
    return tuple(sorted(set(indices), key=id))


def index_sum(expression, indices):
    """Eliminates indices from the free indices of an expression by
    summing over them.  Skips any index that is not a free index of
    the expression."""
    multiindex = tuple(index for index in indices
                       if index in expression.free_indices)
    return IndexSum(expression, multiindex)


def partial_indexed(tensor, indices):
    """Generalised indexing into a tensor by eating shape off the front.
    The number of indices may be less than or equal to the rank of the tensor,
    so the result may have a non-empty shape.

    :arg tensor: tensor-valued GEM expression
    :arg indices: indices, at most as many as the rank of the tensor
    :returns: a potentially tensor-valued expression
    """
    if len(indices) == 0:
        return tensor
    elif len(indices) < len(tensor.shape):
        rank = len(tensor.shape) - len(indices)
        shape_indices = tuple(Index() for i in range(rank))
        return ComponentTensor(
            Indexed(tensor, indices + shape_indices),
            shape_indices)
    elif len(indices) == len(tensor.shape):
        return Indexed(tensor, indices)
    else:
        raise ValueError("More indices than rank!")


def strides_of(shape):
    """Calculate cumulative strides from per-dimension capacities.

    For example:

        [2, 3, 4] ==> [12, 4, 1]

    """
    temp = numpy.flipud(numpy.cumprod(numpy.flipud(list(shape)[1:])))
    return list(temp) + [1]


def decompose_variable_view(expression):
    """Extract information from a shaped node.
       Decompose ComponentTensor + FlexiblyIndexed."""
    if (isinstance(expression, (Variable, Inverse, Solve))):
        variable = expression
        indexes = tuple(Index(extent=extent) for extent in expression.shape)
        dim2idxs = tuple((0, ((index, 1),)) for index in indexes)
    elif (isinstance(expression, ComponentTensor) and
          not isinstance(expression.children[0], FlexiblyIndexed)):
        variable = expression
        indexes = expression.multiindex
        dim2idxs = tuple((0, ((index, 1),)) for index in indexes)
    elif isinstance(expression, ComponentTensor) and isinstance(expression.children[0], FlexiblyIndexed):
        variable = expression.children[0].children[0]
        indexes = expression.multiindex
        dim2idxs = expression.children[0].dim2idxs
    else:
        raise ValueError("Cannot handle {} objects.".format(type(expression).__name__))

    return variable, dim2idxs, indexes


def reshape(expression, *shapes):
    """Reshape a variable (splitting indices only).

    :arg expression: view of a :py:class:`Variable`
    :arg shapes: one shape tuple for each dimension of the variable.
    """
    variable, dim2idxs, indexes = decompose_variable_view(expression)
    assert len(indexes) == len(shapes)
    shape_of = dict(zip(indexes, shapes))

    dim2idxs_ = []
    indices = [[] for _ in range(len(indexes))]
    for offset, idxs in dim2idxs:
        idxs_ = []
        for idx in idxs:
            index, stride = idx
            assert isinstance(index, Index)
            dim = index.extent
            shape = shape_of[index]
            if dim is not None and numpy.prod(shape) != dim:
                raise ValueError("Shape {} does not match extent {}.".format(shape, dim))
            strides = strides_of(shape)
            for extent, stride_ in zip(shape, strides):
                index_ = Index(extent=extent)
                idxs_.append((index_, stride_ * stride))
                indices[indexes.index(index)].append(index_)
        dim2idxs_.append((offset, tuple(idxs_)))

    expr = FlexiblyIndexed(variable, tuple(dim2idxs_))
    return ComponentTensor(expr, tuple(chain.from_iterable(indices)))


def view(expression, *slices):
    """View a part of a shaped object.

    :arg expression: a node that has a shape
    :arg slices: one slice object for each dimension of the expression.
    """
    variable, dim2idxs, indexes = decompose_variable_view(expression)
    assert len(indexes) == len(slices)
    slice_of = dict(zip(indexes, slices))

    dim2idxs_ = []
    indices = [None] * len(slices)
    for offset, idxs in dim2idxs:
        offset_ = offset
        idxs_ = []
        for idx in idxs:
            index, stride = idx
            assert isinstance(index, Index)
            dim = index.extent
            s = slice_of[index]
            start = s.start or 0
            stop = s.stop or dim
            if stop is None:
                raise ValueError("Unknown extent!")
            if dim is not None and stop > dim:
                raise ValueError("Slice exceeds dimension extent!")
            step = s.step or 1
            offset_ += start * stride
            extent = 1 + (stop - start - 1) // step
            index_ = Index(extent=extent)
            indices[indexes.index(index)] = index_
            idxs_.append((index_, step * stride))
        dim2idxs_.append((offset_, tuple(idxs_)))

    expr = FlexiblyIndexed(variable, tuple(dim2idxs_))
    return ComponentTensor(expr, tuple(indices))


# Static one object for quicker constant folding
one = Literal(1)


# Syntax sugar
def indices(n):
    """Make some :class:`Index` objects.

    :arg n: The number of indices to make.
    :returns: A tuple of `n` :class:`Index` objects.
    """
    return tuple(Index() for _ in range(n))


def componentwise(op, *exprs):
    """Apply gem op to exprs component-wise and wrap up in a ComponentTensor.

    :arg op: function that returns a gem Node.
    :arg exprs: expressions to apply op to.
    :raises ValueError: if the expressions have mismatching shapes.
    :returns: New gem Node constructed from op.

    Each expression must either have the same shape, or else be
    scalar. Shaped expressions are indexed, the op is applied to the
    scalar expressions and the result is wrapped up in a ComponentTensor.

    """
    shapes = set(e.shape for e in exprs)
    if len(shapes - {()}) > 1:
        raise ValueError("expressions must have matching shape (or else be scalar)")
    shape = max(shapes)
    i = indices(len(shape))
    exprs = tuple(Indexed(e, i) if e.shape else e for e in exprs)
    return ComponentTensor(op(*exprs), i)


def as_gem(expr):
    """Attempt to convert an expression into GEM of scalar type.

    Parameters
    ----------
    expr : Node or Number
        The expression.

    Returns
    -------
    Node
        A GEM representation of the expression.

    Raises
    ------
    ValueError
        If conversion was not possible.

    """
    if isinstance(expr, Node):
        return expr
    elif isinstance(expr, Number):
        return Literal(expr)
    else:
        raise ValueError("Do not know how to convert %r to GEM" % expr)


def as_gem_uint(expr):
    """Attempt to convert an expression into GEM of uint type.

    Parameters
    ----------
    expr : Node or Integral
        The expression.

    Returns
    -------
    Node
        A GEM representation of the expression.

    Raises
    ------
    ValueError
        If conversion was not possible.

    """
    if isinstance(expr, Node):
        return expr
    elif isinstance(expr, Integral):
        return Literal(expr, dtype=uint_type)
    else:
        raise ValueError("Do not know how to convert %r to GEM" % expr)


def extract_type(expressions, klass):
    """Collects objects of type klass in expressions."""
    return tuple(node for node in traversal(expressions) if isinstance(node, klass))



"""A set of routines implementing various transformations on GEM
expressions."""

from collections import OrderedDict, defaultdict
from functools import singledispatch, partial
from itertools import combinations, permutations, zip_longest
from numbers import Integral

import numpy

from gem.utils import groupby
from gem.node import (Memoizer, MemoizerArg, reuse_if_untouched,
                      reuse_if_untouched_arg, traversal)
from gem.gem import (Node, Failure, Identity, Literal, Zero,
                     Product, Sum, Comparison, Conditional, Division,
                     Index, VariableIndex, Indexed, FlexiblyIndexed,
                     IndexSum, ComponentTensor, ListTensor, Delta,
                     partial_indexed, one)


@singledispatch
def literal_rounding(node, self):
    """Perform FFC rounding of FIAT tabulation matrices on the literals of
    a GEM expression.

    :arg node: root of the expression
    :arg self: function for recursive calls
    """
    raise AssertionError("cannot handle type %s" % type(node))


literal_rounding.register(Node)(reuse_if_untouched)


@literal_rounding.register(Literal)
def literal_rounding_literal(node, self):
    table = node.array
    epsilon = self.epsilon
    # Mimic the rounding applied at COFFEE formatting, which in turn
    # mimics FFC formatting.
    one_decimal = numpy.asarray(numpy.round(table, 1))
    one_decimal[numpy.logical_not(one_decimal)] = 0  # no minus zeros
    return Literal(numpy.where(abs(table - one_decimal) < epsilon, one_decimal, table))


def ffc_rounding(expression, epsilon):
    """Perform FFC rounding of FIAT tabulation matrices on the literals of
    a GEM expression.

    :arg expression: GEM expression
    :arg epsilon: tolerance limit for rounding
    """
    mapper = Memoizer(literal_rounding)
    mapper.epsilon = epsilon
    return mapper(expression)


@singledispatch
def _replace_division(node, self):
    """Replace division with multiplication

    :param node: root of expression
    :param self: function for recursive calls
    """
    raise AssertionError("cannot handle type %s" % type(node))


_replace_division.register(Node)(reuse_if_untouched)


@_replace_division.register(Division)
def _replace_division_division(node, self):
    a, b = node.children
    return Product(self(a), Division(one, self(b)))


def replace_division(expressions):
    """Replace divisions with multiplications in expressions"""
    mapper = Memoizer(_replace_division)
    return list(map(mapper, expressions))


@singledispatch
def replace_indices(node, self, subst):
    """Replace free indices in a GEM expression.

    :arg node: root of the expression
    :arg self: function for recursive calls
    :arg subst: tuple of pairs; each pair is a substitution
                rule with a free index to replace and an index to
                replace with.
    """
    raise AssertionError("cannot handle type %s" % type(node))


replace_indices.register(Node)(reuse_if_untouched_arg)


def _replace_indices_atomic(i, self, subst):
    if isinstance(i, VariableIndex):
        new_expr = self(i.expression, subst)
        return i if new_expr == i.expression else VariableIndex(new_expr)
    else:
        substitute = dict(subst)
        return substitute.get(i, i)


@replace_indices.register(Delta)
def replace_indices_delta(node, self, subst):
    i = _replace_indices_atomic(node.i, self, subst)
    j = _replace_indices_atomic(node.j, self, subst)
    if i == node.i and j == node.j:
        return node
    else:
        return Delta(i, j)


@replace_indices.register(Indexed)
def replace_indices_indexed(node, self, subst):
    child, = node.children
    substitute = dict(subst)
    multiindex = []
    for i in node.multiindex:
        multiindex.append(_replace_indices_atomic(i, self, subst))
    if isinstance(child, ComponentTensor):
        # Indexing into ComponentTensor
        # Inline ComponentTensor and augment the substitution rules
        substitute.update(zip(child.multiindex, multiindex))
        return self(child.children[0], tuple(sorted(substitute.items())))
    else:
        # Replace indices
        new_child = self(child, subst)
        if new_child == child and multiindex == node.multiindex:
            return node
        else:
            return Indexed(new_child, multiindex)


@replace_indices.register(FlexiblyIndexed)
def replace_indices_flexiblyindexed(node, self, subst):
    child, = node.children
    assert not child.free_indices

    dim2idxs = tuple(
        (
            offset if isinstance(offset, Integral) else _replace_indices_atomic(offset, self, subst),
            tuple((_replace_indices_atomic(i, self, subst), s if isinstance(s, Integral) else self(s, subst)) for i, s in idxs)
        )
        for offset, idxs in node.dim2idxs
    )

    if dim2idxs == node.dim2idxs:
        return node
    else:
        return FlexiblyIndexed(child, dim2idxs)


def filtered_replace_indices(node, self, subst):
    """Wrapper for :func:`replace_indices`.  At each call removes
    substitution rules that do not apply."""
    if any(isinstance(k, VariableIndex) for k, _ in subst):
        raise NotImplementedError("Can not replace VariableIndex (will need inverse)")
    filtered_subst = tuple((k, v) for k, v in subst if k in node.free_indices)
    return replace_indices(node, self, filtered_subst)


def remove_componenttensors(expressions):
    """Removes all ComponentTensors in multi-root expression DAG."""
    mapper = MemoizerArg(filtered_replace_indices)
    return [mapper(expression, ()) for expression in expressions]


@singledispatch
def _constant_fold_zero(node, self):
    raise AssertionError("cannot handle type %s" % type(node))


_constant_fold_zero.register(Node)(reuse_if_untouched)


@_constant_fold_zero.register(Literal)
def _constant_fold_zero_literal(node, self):
    if (node.array == 0).all():
        # All zeros, make symbolic zero
        return Zero(node.shape)
    else:
        return node


@_constant_fold_zero.register(ListTensor)
def _constant_fold_zero_listtensor(node, self):
    new_children = list(map(self, node.children))
    if all(isinstance(nc, Zero) for nc in new_children):
        return Zero(node.shape)
    elif all(nc == c for nc, c in zip(new_children, node.children)):
        return node
    else:
        return node.reconstruct(*new_children)


def constant_fold_zero(exprs):
    """Produce symbolic zeros from Literals

    :arg exprs: An iterable of gem expressions.
    :returns: A list of gem expressions where any Literal containing
        only zeros is replaced by symbolic Zero of the appropriate
        shape.

    We need a separate path for ListTensor so that its `reconstruct`
    method will not be called when the new children are `Zero()`s;
    otherwise Literal `0`s would be reintroduced.
    """
    mapper = Memoizer(_constant_fold_zero)
    return [mapper(e) for e in exprs]


def _select_expression(expressions, index):
    """Helper function to select an expression from a list of
    expressions with an index.  This function expect sanitised input,
    one should normally call :py:func:`select_expression` instead.

    :arg expressions: a list of expressions
    :arg index: an index (free, fixed or variable)
    :returns: an expression
    """
    expr = expressions[0]
    if all(e == expr for e in expressions):
        return expr

    types = set(map(type, expressions))
    if types <= {Indexed, Zero}:
        multiindex, = set(e.multiindex for e in expressions if isinstance(e, Indexed))
        # Shape only determined by free indices
        shape = tuple(i.extent for i in multiindex if isinstance(i, Index))

        def child(expression):
            if isinstance(expression, Indexed):
                return expression.children[0]
            elif isinstance(expression, Zero):
                return Zero(shape)
        return Indexed(_select_expression(list(map(child, expressions)), index), multiindex)

    if types <= {Literal, Zero, Failure}:
        return partial_indexed(ListTensor(expressions), (index,))

    if types <= {ComponentTensor, Zero}:
        shape, = set(e.shape for e in expressions)
        multiindex = tuple(Index(extent=d) for d in shape)
        children = remove_componenttensors([Indexed(e, multiindex) for e in expressions])
        return ComponentTensor(_select_expression(children, index), multiindex)

    if len(types) == 1:
        cls, = types
        if cls.__front__ or cls.__back__:
            raise NotImplementedError("How to factorise {} expressions?".format(cls.__name__))
        assert all(len(e.children) == len(expr.children) for e in expressions)
        assert len(expr.children) > 0

        return expr.reconstruct(*[_select_expression(nth_children, index)
                                  for nth_children in zip(*[e.children
                                                            for e in expressions])])

    raise NotImplementedError("No rule for factorising expressions of this kind.")


def select_expression(expressions, index):
    """Select an expression from a list of expressions with an index.
    Semantically equivalent to

        partial_indexed(ListTensor(expressions), (index,))

    but has a much more optimised implementation.

    :arg expressions: a list of expressions of the same shape
    :arg index: an index (free, fixed or variable)
    :returns: an expression of the same shape as the given expressions
    """
    # Check arguments
    shape = expressions[0].shape
    assert all(e.shape == shape for e in expressions)

    # Sanitise input expressions
    alpha = tuple(Index() for s in shape)
    exprs = remove_componenttensors([Indexed(e, alpha) for e in expressions])

    # Factor the expressions recursively and convert result
    selected = _select_expression(exprs, index)
    return ComponentTensor(selected, alpha)


def delta_elimination(sum_indices, factors):
    """IndexSum-Delta cancellation.

    :arg sum_indices: free indices for contractions
    :arg factors: product factors
    :returns: optimised (sum_indices, factors)
    """
    sum_indices = list(sum_indices)  # copy for modification

    def substitute(expression, from_, to_):
        if from_ not in expression.free_indices:
            return expression
        elif isinstance(expression, Delta):
            mapper = MemoizerArg(filtered_replace_indices)
            return mapper(expression, ((from_, to_),))
        else:
            return Indexed(ComponentTensor(expression, (from_,)), (to_,))

    delta_queue = [(f, index)
                   for f in factors if isinstance(f, Delta)
                   for index in (f.i, f.j) if index in sum_indices]
    while delta_queue:
        delta, from_ = delta_queue[0]
        to_, = list({delta.i, delta.j} - {from_})

        sum_indices.remove(from_)

        factors = [substitute(f, from_, to_) for f in factors]

        delta_queue = [(f, index)
                       for f in factors if isinstance(f, Delta)
                       for index in (f.i, f.j) if index in sum_indices]

    return sum_indices, factors


def associate(operator, operands):
    """Apply associativity rules to construct an operation-minimal expression tree.

    For best performance give factors that have different set of free indices.

    :arg operator: associative binary operator
    :arg operands: list of operands

    :returns: (reduced expression, # of floating-point operations)
    """
    if len(operands) > 32:
        # O(N^3) algorithm
        raise NotImplementedError("Not expected such a complicated expression!")

    def count(pair):
        """Operation count to reduce a pair of GEM expressions"""
        a, b = pair
        extents = [i.extent for i in set().union(a.free_indices, b.free_indices)]
        return numpy.prod(extents, dtype=int)

    flops = 0
    while len(operands) > 1:
        # Greedy algorithm: choose a pair of operands that are the
        # cheapest to reduce.
        a, b = min(combinations(operands, 2), key=count)
        flops += count((a, b))
        # Remove chosen factors, append their product
        operands.remove(a)
        operands.remove(b)
        operands.append(operator(a, b))
    result, = operands
    return result, flops


def sum_factorise(sum_indices, factors):
    """Optimise a tensor product through sum factorisation.

    :arg sum_indices: free indices for contractions
    :arg factors: product factors
    :returns: optimised GEM expression
    """
    if len(factors) == 0 and len(sum_indices) == 0:
        # Empty product
        return one

    if len(sum_indices) > 6:
        raise NotImplementedError("Too many indices for sum factorisation!")

    # Form groups by free indices
    groups = groupby(factors, key=lambda f: f.free_indices)
    groups = [Product(*terms) for _, terms in groups]

    # Sum factorisation
    expression = None
    best_flops = numpy.inf

    # Consider all orderings of contraction indices
    for ordering in permutations(sum_indices):
        terms = groups[:]
        flops = 0
        # Apply contraction index by index
        for sum_index in ordering:
            # Select terms that need to be part of the contraction
            contract = [t for t in terms if sum_index in t.free_indices]
            deferred = [t for t in terms if sum_index not in t.free_indices]

            # Optimise associativity
            product, flops_ = associate(Product, contract)
            term = IndexSum(product, (sum_index,))
            flops += flops_ + numpy.prod([i.extent for i in product.free_indices], dtype=int)

            # Replace the contracted terms with the result of the
            # contraction.
            terms = deferred + [term]

        # If some contraction indices were independent, then we may
        # still have several terms at this point.
        expr, flops_ = associate(Product, terms)
        flops += flops_

        if flops < best_flops:
            expression = expr
            best_flops = flops

    return expression


def make_sum(summands):
    """Constructs an operation-minimal sum of GEM expressions."""
    groups = groupby(summands, key=lambda f: f.free_indices)
    summands = [Sum(*terms) for _, terms in groups]
    result, flops = associate(Sum, summands)
    return result


def make_product(factors, sum_indices=()):
    """Constructs an operation-minimal (tensor) product of GEM expressions."""
    return sum_factorise(sum_indices, factors)


def make_rename_map():
    """Creates an rename map for reusing the same index renames."""
    return defaultdict(Index)


def make_renamer(rename_map):
    r"""Creates a function for renaming indices when expanding products of
    IndexSums, i.e. applying to following rule:

        (\sum_i a_i)*(\sum_i b_i) ===> \sum_{i,i'} a_i*b_{i'}

    :arg rename_map: An rename map for renaming indices the same way
                     as functions returned by other calls of this
                     function.
    :returns: A function that takes an iterable of indices to rename,
              and returns (renamed indices, applier), where applier is
              a function that remap the free indices of GEM
              expressions from the old to the new indices.
    """
    def _renamer(rename_map, current_set, incoming):
        renamed = []
        renames = []
        for i in incoming:
            j = i
            while j in current_set:
                j = rename_map[j]
            current_set.add(j)
            renamed.append(j)
            if i != j:
                renames.append((i, j))

        if renames:
            def applier(expr):
                pairs = [(i, j) for i, j in renames if i in expr.free_indices]
                if pairs:
                    current, renamed = zip(*pairs)
                    return Indexed(ComponentTensor(expr, current), renamed)
                else:
                    return expr
        else:
            applier = lambda expr: expr

        return tuple(renamed), applier
    return partial(_renamer, rename_map, set())


def traverse_product(expression, stop_at=None, rename_map=None):
    """Traverses a product tree and collects factors, also descending into
    tensor contractions (IndexSum).  The nominators of divisions are
    also broken up, but not the denominators.

    :arg expression: a GEM expression
    :arg stop_at: Optional predicate on GEM expressions.  If specified
                  and returns true for some subexpression, that
                  subexpression is not broken into further factors
                  even if it is a product-like expression.
    :arg rename_map: an rename map for consistent index renaming
    :returns: (sum_indices, terms)
              - sum_indices: list of indices to sum over
              - terms: list of product terms
    """
    if rename_map is None:
        rename_map = make_rename_map()
    renamer = make_renamer(rename_map)

    sum_indices = []
    terms = []

    stack = [expression]
    while stack:
        expr = stack.pop()
        if stop_at is not None and stop_at(expr):
            terms.append(expr)
        elif isinstance(expr, IndexSum):
            indices, applier = renamer(expr.multiindex)
            sum_indices.extend(indices)
            stack.extend(remove_componenttensors(map(applier, expr.children)))
        elif isinstance(expr, Product):
            stack.extend(reversed(expr.children))
        elif isinstance(expr, Division):
            # Break up products in the dividend, but not in divisor.
            dividend, divisor = expr.children
            if dividend == one:
                terms.append(expr)
            else:
                stack.append(Division(one, divisor))
                stack.append(dividend)
        else:
            terms.append(expr)

    return sum_indices, terms


def traverse_sum(expression, stop_at=None):
    """Traverses a summation tree and collects summands.

    :arg expression: a GEM expression
    :arg stop_at: Optional predicate on GEM expressions.  If specified
                  and returns true for some subexpression, that
                  subexpression is not broken into further summands
                  even if it is an addition.
    :returns: list of summand expressions
    """
    stack = [expression]
    result = []
    while stack:
        expr = stack.pop()
        if stop_at is not None and stop_at(expr):
            result.append(expr)
        elif isinstance(expr, Sum):
            stack.extend(reversed(expr.children))
        else:
            result.append(expr)
    return result


def contraction(expression, ignore=None):
    """Optimise the contractions of the tensor product at the root of
    the expression, including:

    - IndexSum-Delta cancellation
    - Sum factorisation

    :arg ignore: Optional set of indices to ignore when applying sum
        factorisation (otherwise all summation indices will be
        considered). Use this if your expression has many contraction
        indices.

    This routine was designed with finite element coefficient
    evaluation in mind.
    """
    # Eliminate annoying ComponentTensors
    expression, = remove_componenttensors([expression])

    # Flatten product tree, eliminate deltas, sum factorise
    def rebuild(expression):
        sum_indices, factors = delta_elimination(*traverse_product(expression))
        factors = remove_componenttensors(factors)
        if ignore is not None:
            # TODO: This is a really blunt instrument and one might
            # plausibly want the ignored indices to be contracted on
            # the inside rather than the outside.
            extra = tuple(i for i in sum_indices if i in ignore)
            to_factor = tuple(i for i in sum_indices if i not in ignore)
            return IndexSum(sum_factorise(to_factor, factors), extra)
        else:
            return sum_factorise(sum_indices, factors)

    # Sometimes the value shape is composed as a ListTensor, which
    # could get in the way of decomposing factors.  In particular,
    # this is the case for H(div) and H(curl) conforming tensor
    # product elements.  So if ListTensors are used, they are pulled
    # out to be outermost, so we can straightforwardly factorise each
    # of its entries.
    lt_fis = OrderedDict()  # ListTensor free indices
    for node in traversal((expression,)):
        if isinstance(node, Indexed):
            child, = node.children
            if isinstance(child, ListTensor):
                lt_fis.update(zip_longest(node.multiindex, ()))
    lt_fis = tuple(index for index in lt_fis if index in expression.free_indices)

    if lt_fis:
        # Rebuild each split component
        tensor = ComponentTensor(expression, lt_fis)
        entries = [Indexed(tensor, zeta) for zeta in numpy.ndindex(tensor.shape)]
        entries = remove_componenttensors(entries)
        return Indexed(ListTensor(
            numpy.array(list(map(rebuild, entries))).reshape(tensor.shape)
        ), lt_fis)
    else:
        # Rebuild whole expression at once
        return rebuild(expression)


@singledispatch
def _replace_delta(node, self):
    raise AssertionError("cannot handle type %s" % type(node))


_replace_delta.register(Node)(reuse_if_untouched)


@_replace_delta.register(Delta)
def _replace_delta_delta(node, self):
    i, j = node.i, node.j

    if isinstance(i, Index) or isinstance(j, Index):
        if isinstance(i, Index) and isinstance(j, Index):
            assert i.extent == j.extent
        if isinstance(i, Index):
            assert i.extent is not None
            size = i.extent
        if isinstance(j, Index):
            assert j.extent is not None
            size = j.extent
        return Indexed(Identity(size), (i, j))
    else:
        def expression(index):
            if isinstance(index, Integral):
                return Literal(index)
            elif isinstance(index, VariableIndex):
                return index.expression
            else:
                raise ValueError("Cannot convert running index to expression.")
        e_i = expression(i)
        e_j = expression(j)
        return Conditional(Comparison("==", e_i, e_j), one, Zero())


def replace_delta(expressions):
    """Lowers all Deltas in a multi-root expression DAG."""
    mapper = Memoizer(_replace_delta)
    return list(map(mapper, expressions))


@singledispatch
def _unroll_indexsum(node, self):
    """Unrolls IndexSums below a certain extent.

    :arg node: root of the expression
    :arg self: function for recursive calls
    """
    raise AssertionError("cannot handle type %s" % type(node))


_unroll_indexsum.register(Node)(reuse_if_untouched)


@_unroll_indexsum.register(IndexSum)  # noqa
def _(node, self):
    unroll = tuple(filter(self.predicate, node.multiindex))
    if unroll:
        # Unrolling
        summand = self(node.children[0])
        shape = tuple(index.extent for index in unroll)
        unrolled = Sum(*(Indexed(ComponentTensor(summand, unroll), alpha)
                         for alpha in numpy.ndindex(shape)))
        return IndexSum(unrolled, tuple(index for index in node.multiindex
                                        if index not in unroll))
    else:
        return reuse_if_untouched(node, self)


def unroll_indexsum(expressions, predicate):
    """Unrolls IndexSums below a specified extent.

    :arg expressions: list of expression DAGs
    :arg predicate: a predicate function on :py:class:`Index` objects
                    that tells whether to unroll a particular index
    :returns: list of expression DAGs with some unrolled IndexSums
    """
    mapper = Memoizer(_unroll_indexsum)
    mapper.predicate = predicate
    return list(map(mapper, expressions))


def aggressive_unroll(expression):
    """Aggressively unrolls all loop structures."""
    # Unroll expression shape
    if expression.shape:
        tensor = numpy.empty(expression.shape, dtype=object)
        for alpha in numpy.ndindex(expression.shape):
            tensor[alpha] = Indexed(expression, alpha)
        expression, = remove_componenttensors((ListTensor(tensor),))

    # Unroll summation
    expression, = unroll_indexsum((expression,), predicate=lambda index: True)
    expression, = remove_componenttensors((expression,))
    return expression



# Copyright (C) 2008 Robert C. Kirby (Texas Tech University)
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from FIAT import finite_element, polynomial_set, dual_set
from FIAT.check_format_variant import check_format_variant
from FIAT.functional import (PointEvaluation, PointDerivative, PointNormalDerivative,
                             IntegralMoment,
                             IntegralMomentOfNormalDerivative)
from FIAT.jacobi import eval_jacobi_batch, eval_jacobi_deriv_batch
from FIAT.quadrature import FacetQuadratureRule
from FIAT.quadrature_schemes import create_quadrature
from FIAT.reference_element import TRIANGLE, ufc_simplex


class ArgyrisDualSet(dual_set.DualSet):
    def __init__(self, ref_el, degree, variant, interpolant_deg):
        if ref_el.get_shape() != TRIANGLE:
            raise ValueError("Argyris only defined on triangles")

        top = ref_el.get_topology()
        sd = ref_el.get_spatial_dimension()
        entity_ids = {dim: {entity: [] for entity in sorted(top[dim])} for dim in sorted(top)}
        nodes = []

        # get second order jet at each vertex
        verts = ref_el.get_vertices()
        alphas = [(1, 0), (0, 1), (2, 0), (1, 1), (0, 2)]
        for v in sorted(top[0]):
            cur = len(nodes)
            nodes.append(PointEvaluation(ref_el, verts[v]))
            nodes.extend(PointDerivative(ref_el, verts[v], alpha) for alpha in alphas)
            entity_ids[0][v] = list(range(cur, len(nodes)))

        if variant == "integral":
            # edge dofs
            k = degree - 5
            rline = ufc_simplex(1)
            Q = create_quadrature(rline, interpolant_deg+k-1)
            x = 2.0 * Q.get_points() - 1.0
            phis = eval_jacobi_batch(2, 2, k, x)
            dphis = eval_jacobi_deriv_batch(2, 2, k, x)
            for e in sorted(top[1]):
                Q_mapped = FacetQuadratureRule(ref_el, 1, e, Q)
                scale = 2 / Q_mapped.jacobian_determinant()
                cur = len(nodes)
                nodes.extend(IntegralMomentOfNormalDerivative(ref_el, e, Q, phi) for phi in phis)
                nodes.extend(IntegralMoment(ref_el, Q_mapped, dphi * scale) for dphi in dphis[1:])
                entity_ids[1][e].extend(range(cur, len(nodes)))

            # interior dofs
            q = degree - 6
            if q >= 0:
                cell = ref_el.construct_subelement(sd)
                Q_ref = create_quadrature(cell, interpolant_deg + q)
                Pq = polynomial_set.ONPolynomialSet(cell, q, scale=1)
                Phis = Pq.tabulate(Q_ref.get_points())[(0,) * sd]
                for entity in sorted(top[sd]):
                    Q = FacetQuadratureRule(ref_el, sd, entity, Q_ref)
                    scale = 1 / Q.jacobian_determinant()
                    phis = scale * Phis
                    cur = len(nodes)
                    nodes.extend(IntegralMoment(ref_el, Q, phi) for phi in phis)
                    entity_ids[sd][entity] = list(range(cur, len(nodes)))

        elif variant == "point":
            # edge dofs
            for e in sorted(top[1]):
                cur = len(nodes)
                # normal derivatives at degree - 4 points on each edge
                ndpts = ref_el.make_points(1, e, degree - 3)
                nodes.extend(PointNormalDerivative(ref_el, e, pt) for pt in ndpts)

                # point value at degree - 5 points on each edge
                ptvalpts = ref_el.make_points(1, e, degree - 4)
                nodes.extend(PointEvaluation(ref_el, pt) for pt in ptvalpts)
                entity_ids[1][e] = list(range(cur, len(nodes)))

            # interior dofs
            if degree > 5:
                cur = len(nodes)
                for entity in sorted(top[sd]):
                    internalpts = ref_el.make_points(sd, entity, degree - 3)
                    nodes.extend(PointEvaluation(ref_el, pt) for pt in internalpts)
                    entity_ids[sd][entity] = list(range(cur, len(nodes)))
        else:
            raise ValueError("Invalid variant for Argyris")
        super().__init__(nodes, ref_el, entity_ids)


class Argyris(finite_element.CiarletElement):
    """
    The Argyris finite element.

    :arg ref_el: The reference element.
    :arg degree: The degree.
    :arg variant: optional variant specifying the types of nodes.

    variant can be chosen from ["point", "integral", "integral(q)"]
    "point" -> dofs are evaluated by point evaluation.
    "integral" -> dofs are evaluated by quadrature rules with the minimum
    degree required for unisolvence.
    "integral(q)" -> dofs are evaluated by quadrature rules with the minimum
    degree required for unisolvence plus q.
    """

    def __init__(self, ref_el, degree=5, variant=None):

        splitting, variant, interpolant_deg = check_format_variant(variant, degree)
        if splitting is not None:
            raise NotImplementedError(f"{type(self).__name__} is not implemented as a macroelement.")

        poly_set = polynomial_set.ONPolynomialSet(ref_el, degree, variant="bubble")
        dual = ArgyrisDualSet(ref_el, degree, variant, interpolant_deg)
        super().__init__(poly_set, dual, degree)



import os
import sys
import subprocess
import textwrap

import ufl
import finat.ufl


def test_same_hash():
    """ The same element created twice should have the same hash.
    """
    cg = finat.ufl.finiteelement.FiniteElement("Lagrange", ufl.cell.Cell("triangle"), 1)
    same_cg = finat.ufl.finiteelement.FiniteElement("Lagrange", ufl.cell.Cell("triangle"), 1)
    assert hash(cg) == hash(same_cg)


def test_different_hash():
    """ Two different elements should have different hashes.
    """
    cg = finat.ufl.finiteelement.FiniteElement("Lagrange", ufl.cell.Cell("triangle"), 1)
    dg = finat.ufl.finiteelement.FiniteElement("DG", ufl.cell.Cell("triangle"), 2)
    assert hash(cg) != hash(dg)


def test_variant_hashes_different():
    """ Different variants of the same element should have different hashes.
    """
    dg = finat.ufl.finiteelement.FiniteElement("DG", ufl.cell.Cell("triangle"), 2)
    dg_gll = finat.ufl.finiteelement.FiniteElement("DG", ufl.cell.Cell("triangle"), 2, variant="gll")
    assert hash(dg) != hash(dg_gll)


def test_persistent_hash(tmp_path):
    """ Hashes should be the same across Python invocations.
    """
    filename = "print_hash.py"
    code = textwrap.dedent("""\
        import ufl
        import finat.ufl

        dg = finat.ufl.finiteelement.FiniteElement("RT", ufl.cell.Cell("triangle"), 1)
        print(hash(dg))
        """)
    filepath = tmp_path.joinpath(filename)
    with open(filepath, "w") as fh:
        fh.write(code)

    output1 = subprocess.run([sys.executable, filepath], capture_output=True)
    assert output1.returncode == os.EX_OK
    output2 = subprocess.run([sys.executable, filepath], capture_output=True)
    assert output2.returncode == os.EX_OK
    assert output1.stdout == output2.stdout



"""Doc."""
# Copyright (C) 2008-2016 Andrew T. T. McRae
#
# This file was originally part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Massimiliano Leoni, 2016
# Modified by Matthew Scroggs, 2023

from finat.ufl.finiteelementbase import FiniteElementBase
from ufl.sobolevspace import L2, SobolevSpace
from ufl.sobolevspace import HCurl as HCurlSobolevSpace
from ufl.sobolevspace import HDiv as HDivSobolevSpace


class CallableSobolevSpace(SobolevSpace):
    """A Sobolev space that can be called to create HDiv and HCurl elements."""

    def __init__(self, name, parents=None):
        super().__init__(name, parents)

    def __call__(self, element):
        """Syntax shortcut to create a HDivElement or HCurlElement."""
        if self.name == "HDiv":
            return HDivElement(element)
        elif self.name == "HCurl":
            return HCurlElement(element)
        raise NotImplementedError(
            "SobolevSpace has no call operator (only the specific HDiv and HCurl instances)."
        )


HCurl = CallableSobolevSpace(HCurlSobolevSpace.name, HCurlSobolevSpace.parents)
HDiv = CallableSobolevSpace(HDivSobolevSpace.name, HDivSobolevSpace.parents)


class HDivElement(FiniteElementBase):
    """A div-conforming version of an outer product element, assuming this makes mathematical sense."""
    __slots__ = ("_element", )

    def __init__(self, element):
        """Doc."""
        self._element = element

        family = "TensorProductElement"
        cell = element.cell
        degree = element.degree()
        quad_scheme = element.quadrature_scheme()
        reference_value_shape = (element.cell.topological_dimension(),)

        # Skipping TensorProductElement constructor! Bad code smell, refactor to avoid this non-inheritance somehow.
        FiniteElementBase.__init__(self, family, cell, degree,
                                   quad_scheme, reference_value_shape)

    def __repr__(self):
        """Doc."""
        return f"HDivElement({repr(self._element)})"

    def mapping(self):
        """Doc."""
        return "contravariant Piola"

    @property
    def sobolev_space(self):
        """Return the underlying Sobolev space."""
        return HDivSobolevSpace

    def reconstruct(self, **kwargs):
        """Doc."""
        return HDivElement(self._element.reconstruct(**kwargs))

    def variant(self):
        """Doc."""
        return self._element.variant()

    def __str__(self):
        """Doc."""
        return f"HDivElement({repr(self._element)})"

    def shortstr(self):
        """Format as string for pretty printing."""
        return f"HDivElement({self._element.shortstr()})"

    @property
    def embedded_subdegree(self):
        """Return embedded subdegree."""
        return self._element.embedded_subdegree

    @property
    def embedded_superdegree(self):
        """Return embedded superdegree."""
        return self._element.embedded_superdegree


class HCurlElement(FiniteElementBase):
    """A curl-conforming version of an outer product element, assuming this makes mathematical sense."""
    __slots__ = ("_element",)

    def __init__(self, element):
        """Doc."""
        self._element = element

        family = "TensorProductElement"
        cell = element.cell
        degree = element.degree()
        quad_scheme = element.quadrature_scheme()
        cell = element.cell
        reference_value_shape = (cell.topological_dimension(),)  # TODO: Is this right?
        # Skipping TensorProductElement constructor! Bad code smell,
        # refactor to avoid this non-inheritance somehow.
        FiniteElementBase.__init__(self, family, cell, degree, quad_scheme,
                                   reference_value_shape)

    def __repr__(self):
        """Doc."""
        return f"HCurlElement({repr(self._element)})"

    def mapping(self):
        """Doc."""
        return "covariant Piola"

    @property
    def sobolev_space(self):
        """Return the underlying Sobolev space."""
        return HCurlSobolevSpace

    def reconstruct(self, **kwargs):
        """Doc."""
        return HCurlElement(self._element.reconstruct(**kwargs))

    def variant(self):
        """Doc."""
        return self._element.variant()

    def __str__(self):
        """Doc."""
        return f"HCurlElement({repr(self._element)})"

    def shortstr(self):
        """Format as string for pretty printing."""
        return f"HCurlElement({self._element.shortstr()})"

    @property
    def embedded_subdegree(self):
        """Return embedded subdegree."""
        return self._element.embedded_subdegree

    @property
    def embedded_superdegree(self):
        """Return embedded superdegree."""
        return self._element.embedded_superdegree


class WithMapping(FiniteElementBase):
    """Specify an alternative mapping for the wrappee.

    For example,
    to use identity mapping instead of Piola map with an element E,
    write
    remapped = WithMapping(E, "identity")
    """

    def __init__(self, wrapee, mapping):
        """Doc."""
        if mapping == "symmetries":
            raise ValueError("Can't change mapping to 'symmetries'")
        self._mapping = mapping
        self.wrapee = wrapee

    def __getattr__(self, attr):
        """Doc."""
        try:
            return getattr(self.wrapee, attr)
        except AttributeError:
            raise AttributeError("'%s' object has no attribute '%s'" %
                                 (type(self).__name__, attr))

    def __repr__(self):
        """Doc."""
        return f"WithMapping({repr(self.wrapee)}, '{self._mapping}')"

    def value_shape(self, domain):
        """Doc."""
        gdim = domain.geometric_dimension()
        mapping = self.mapping()
        if mapping in {"covariant Piola", "contravariant Piola"}:
            return (gdim,)
        elif mapping in {"double covariant Piola", "double contravariant Piola"}:
            return (gdim, gdim)
        else:
            return self.wrapee.value_shape(domain)

    @property
    def reference_value_shape(self):
        """Doc."""
        tdim = self.cell.topological_dimension()
        mapping = self.mapping()
        if mapping in {"covariant Piola", "contravariant Piola"}:
            return (tdim,)
        elif mapping in {"double covariant Piola", "double contravariant Piola"}:
            return (tdim, tdim)
        else:
            return self.wrapee.reference_value_shape

    def mapping(self):
        """Doc."""
        return self._mapping

    @property
    def sobolev_space(self):
        """Return the underlying Sobolev space."""
        if self.wrapee.mapping() == self.mapping():
            return self.wrapee.sobolev_space
        else:
            return L2

    def reconstruct(self, **kwargs):
        """Doc."""
        mapping = kwargs.pop("mapping", self._mapping)
        wrapee = self.wrapee.reconstruct(**kwargs)
        return type(self)(wrapee, mapping)

    def variant(self):
        """Doc."""
        return self.wrapee.variant()

    def __str__(self):
        """Doc."""
        return f"WithMapping({repr(self.wrapee)}, {self._mapping})"

    def shortstr(self):
        """Doc."""
        return f"WithMapping({self.wrapee.shortstr()}, {self._mapping})"

    @property
    def embedded_subdegree(self):
        """Return embedded subdegree."""
        return self._element.embedded_subdegree

    @property
    def embedded_superdegree(self):
        """Return embedded superdegree."""
        return self._element.embedded_superdegree



# Copyright (C) 2024 Pablo Brubeck
#
# This file is part of FIAT.
#
# FIAT is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FIAT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FIAT. If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy
import sympy

from FIAT import expansions, polynomial_set, reference_element
from FIAT.quadrature_schemes import create_quadrature
from itertools import chain


@pytest.fixture(params=(1, 2, 3))
def cell(request):
    dim = request.param
    return reference_element.default_simplex(dim)


@pytest.mark.parametrize("degree", [10])
def test_expansion_values(cell, degree):
    dim = cell.get_spatial_dimension()
    U = expansions.ExpansionSet(cell)
    dpoints = []
    rpoints = []

    numpyoints = 4
    interior = 1
    for alpha in reference_element.lattice_iter(interior, numpyoints+1-interior, dim):
        dpoints.append(tuple(2*numpy.array(alpha, dtype="d")/numpyoints-1))
        rpoints.append(tuple(2*sympy.Rational(a, numpyoints)-1 for a in alpha))

    Uvals = U.tabulate(degree, dpoints)
    idx = (lambda p: p, expansions.morton_index2, expansions.morton_index3)[dim-1]
    eta = sympy.DeferredVector("eta")
    half = sympy.Rational(1, 2)

    def duffy_coords(pt):
        if len(pt) == 1:
            return pt
        elif len(pt) == 2:
            eta0 = 2 * (1 + pt[0]) / (1 - pt[1]) - 1
            eta1 = pt[1]
            return eta0, eta1
        else:
            eta0 = 2 * (1 + pt[0]) / (-pt[1] - pt[2]) - 1
            eta1 = 2 * (1 + pt[1]) / (1 - pt[2]) - 1
            eta2 = pt[2]
            return eta0, eta1, eta2

    def basis(dim, p, q=0, r=0):
        if dim >= 1:
            f = sympy.jacobi(p, 0, 0, eta[0])
            f *= sympy.sqrt(half + p)
        if dim >= 2:
            f *= sympy.jacobi(q, 2*p+1, 0, eta[1]) * ((1 - eta[1])/2) ** p
            f *= sympy.sqrt(1 + p + q)
        if dim >= 3:
            f *= sympy.jacobi(r, 2*p+2*q+2, 0, eta[2]) * ((1 - eta[2])/2) ** (p+q)
            f *= sympy.sqrt(1 + half + p + q + r)
        return f

    def eval_basis(f, pt):
        return float(f.subs(dict(zip(eta, duffy_coords(pt)))))

    for i in range(degree + 1):
        for indices in polynomial_set.mis(dim, i):
            phi = basis(dim, *indices)
            exact = numpy.array([eval_basis(phi, r) for r in rpoints])
            uh = Uvals[idx(*indices)]
            assert numpy.allclose(uh, exact, atol=1E-14)


@pytest.mark.parametrize("degree", [10])
def test_expansion_orthonormality(cell, degree):
    U = expansions.ExpansionSet(cell)
    rule = create_quadrature(cell, 2*degree)
    phi = U.tabulate(degree, rule.pts)
    qwts = rule.get_weights()
    results = numpy.dot(numpy.multiply(phi, qwts), phi.T)
    assert numpy.allclose(results, numpy.diag(numpy.diag(results)))
    assert numpy.allclose(numpy.diag(results), 1.0)


@pytest.mark.parametrize("degree", [10])
def test_bubble_duality(cell, degree):
    sd = cell.get_spatial_dimension()
    B = polynomial_set.make_bubbles(cell, degree)

    Q = create_quadrature(cell, 2*B.degree - sd - 1)
    qpts, qwts = Q.get_points(), Q.get_weights()
    phi = B.tabulate(qpts)[(0,) * sd]
    phi_dual = phi / abs(phi[0])
    scale = 2 ** sd
    results = scale * numpy.dot(numpy.multiply(phi_dual, qwts), phi.T)
    assert numpy.allclose(results, numpy.diag(numpy.diag(results)))
    assert numpy.allclose(numpy.diag(results), 1.0)


@pytest.mark.parametrize("degree", [10])
def test_union_of_polysets(cell, degree):
    """ demonstrates that polysets don't need to have the same degree for union
    using RT space as an example"""

    sd = cell.get_spatial_dimension()
    k = degree
    vecPk = polynomial_set.ONPolynomialSet(cell, degree, (sd,))

    vec_Pkp1 = polynomial_set.ONPolynomialSet(cell, k + 1, (sd,), scale="orthonormal")

    dimPkp1 = expansions.polynomial_dimension(cell, k + 1)
    dimPk = expansions.polynomial_dimension(cell, k)
    dimPkm1 = expansions.polynomial_dimension(cell, k - 1)

    vec_Pk_indices = list(chain(*(range(i * dimPkp1, i * dimPkp1 + dimPk)
                                  for i in range(sd))))
    vec_Pk_from_Pkp1 = vec_Pkp1.take(vec_Pk_indices)

    Pkp1 = polynomial_set.ONPolynomialSet(cell, k + 1, scale="orthonormal")
    PkH = Pkp1.take(list(range(dimPkm1, dimPk)))

    Q = create_quadrature(cell, 2 * (k + 1))
    Qpts, Qwts = Q.get_points(), Q.get_weights()

    PkH_at_Qpts = PkH.tabulate(Qpts)[(0,) * sd]
    Pkp1_at_Qpts = Pkp1.tabulate(Qpts)[(0,) * sd]
    x = Qpts.T
    PkHx_at_Qpts = PkH_at_Qpts[:, None, :] * x[None, :, :]
    PkHx_coeffs = numpy.dot(numpy.multiply(PkHx_at_Qpts, Qwts), Pkp1_at_Qpts.T)
    PkHx = polynomial_set.PolynomialSet(cell, k, k + 1, vec_Pkp1.get_expansion_set(), PkHx_coeffs)

    same_deg = polynomial_set.polynomial_set_union_normalized(vec_Pk_from_Pkp1, PkHx)
    different_deg = polynomial_set.polynomial_set_union_normalized(vecPk, PkHx)

    Q = create_quadrature(cell, 2*(degree))
    Qpts, _ = Q.get_points(), Q.get_weights()
    same_vals = same_deg.tabulate(Qpts)[(0,) * sd]
    diff_vals = different_deg.tabulate(Qpts)[(0,) * sd]
    assert numpy.allclose(same_vals - diff_vals, 0)



import pytest
import numpy

from FIAT import Argyris, HsiehCloughTocher
from FIAT.jacobi import eval_jacobi_batch, eval_jacobi_deriv_batch
from FIAT.quadrature import FacetQuadratureRule
from FIAT.quadrature_schemes import create_quadrature
from FIAT.reference_element import ufc_simplex


@pytest.fixture
def cell():
    return ufc_simplex(2)


def directional_derivative(direction, tab):
    return sum(direction[alpha.index(1)] * tab[alpha]
               for alpha in tab if sum(alpha) == 1)


def inner(u, v, wts):
    return numpy.dot(numpy.multiply(u, wts), v.T)


@pytest.mark.parametrize("family, degree", [
    *((Argyris, k) for k in range(6, 10)),
    *((HsiehCloughTocher, k) for k in range(4, 8))])
def test_argyris_basis_functions(cell, family, degree):
    fe = family(cell, degree)

    ref_el = fe.get_reference_element()
    sd = ref_el.get_spatial_dimension()
    top = ref_el.get_topology()
    entity_ids = fe.entity_dofs()

    degree = fe.degree()
    lowest_p = 5 if isinstance(fe, Argyris) else 3
    a = (lowest_p - 1) // 2
    q = degree - lowest_p

    rline = ufc_simplex(1)
    Qref = create_quadrature(rline, 2*degree)
    xref = 2.0 * Qref.get_points() - 1

    ell_at_qpts = eval_jacobi_batch(0, 0, degree, xref)
    P_at_qpts = eval_jacobi_batch(a, a, degree, xref)
    DP_at_qpts = eval_jacobi_deriv_batch(a, a, degree, xref)

    for e in top[1]:
        n = ref_el.compute_normal(e)
        t = ref_el.compute_edge_tangent(e)
        Q = FacetQuadratureRule(ref_el, 1, e, Qref)
        qpts, qwts = Q.get_points(), Q.get_weights()

        ids = entity_ids[1][e]
        ids1 = ids[1:-q]
        ids0 = ids[-q:]

        tab = fe.tabulate(1, qpts)
        phi = tab[(0,) * sd]
        phi_n = directional_derivative(n, tab)
        phi_t = directional_derivative(t, tab)

        # Test that normal derivative moment bfs have vanishing trace
        assert numpy.allclose(phi[ids[:-q]], 0)

        # Test that trace moment bfs have vanishing normal derivative
        assert numpy.allclose(phi_n[ids0], 0)

        # Test that facet bfs are hierarchical
        coeffs = inner(ell_at_qpts[1+lowest_p:], phi[ids0], qwts)
        assert numpy.allclose(coeffs, numpy.triu(coeffs))
        coeffs = inner(ell_at_qpts[1+lowest_p:], phi_n[ids1], qwts)
        assert numpy.allclose(coeffs, numpy.triu(coeffs))

        # Test duality of normal derivative moments
        coeffs = inner(P_at_qpts[1:], phi_n[ids1], qwts)
        assert numpy.allclose(coeffs[q:], 0)
        assert numpy.allclose(coeffs[:q], numpy.diag(numpy.diag(coeffs[:q])))

        # Test duality of trace moments
        coeffs = inner(DP_at_qpts[1:], phi[ids0], qwts)
        assert numpy.allclose(coeffs[q:], 0)
        assert numpy.allclose(coeffs[:q], numpy.diag(numpy.diag(coeffs[:q])))

        # Test the integration by parts property arising from the choice
        # of normal derivative and trace moments DOFs.
        # The normal derivative of the normal derviative bfs must be equal
        # to minus the tangential derivative of the trace moment bfs
        assert numpy.allclose(numpy.dot((phi_n[ids1] + phi_t[ids0])**2, qwts), 0)



"""Element."""
# -*- coding: utf-8 -*-
# Copyright (C) 2014 Andrew T. T. McRae
#
# This file was originally part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Massimiliano Leoni, 2016
# Modified by Matthew Scroggs, 2023

from finat.ufl.finiteelementbase import FiniteElementBase
from ufl.sobolevspace import L2


class BrokenElement(FiniteElementBase):
    """The discontinuous version of an existing Finite Element space."""
    def __init__(self, element):
        """Init."""
        self._element = element

        family = "BrokenElement"
        cell = element.cell
        degree = element.degree()
        quad_scheme = element.quadrature_scheme()
        reference_value_shape = element.reference_value_shape
        FiniteElementBase.__init__(self, family, cell, degree,
                                   quad_scheme, reference_value_shape)

    def __repr__(self):
        """Doc."""
        return f"BrokenElement({repr(self._element)})"

    def mapping(self):
        """Doc."""
        return self._element.mapping()

    @property
    def sobolev_space(self):
        """Return the underlying Sobolev space."""
        return L2

    def reconstruct(self, **kwargs):
        """Doc."""
        return BrokenElement(self._element.reconstruct(**kwargs))

    def __str__(self):
        """Doc."""
        return f"BrokenElement({repr(self._element)})"

    def shortstr(self):
        """Format as string for pretty printing."""
        return f"BrokenElement({repr(self._element)})"

    @property
    def embedded_subdegree(self):
        """Return embedded subdegree."""
        return self._element.embedded_subdegree

    @property
    def embedded_superdegree(self):
        """Return embedded superdegree."""
        return self._element.embedded_superdegree



import itertools
import math
from collections.abc import Sequence
import numpy as np


class Orientation(object):
    """Base class representing unsigned integer orientations.

    Orientations represented by this class are to be consistent with those used in
    `make_entity_permutations_simplex` and `make_entity_permutations_tensorproduct`.

    """

    def __floordiv__(self, other):
        raise NotImplementedError("Subclass must implement __floordiv__")

    def __rfloordiv__(self, other):
        raise NotImplementedError("Subclass must implement __rfloordiv__")

    def __mod__(self, other):
        raise NotImplementedError("Subclass must implement __mod__")

    def __rmod__(self, other):
        raise NotImplementedError("Subclass must implement __rmod__")


def make_entity_permutations_simplex(dim, npoints):
    r"""Make orientation-permutation map for the given
    simplex dimension, dim, and the number of points along
    each axis

    As an example, we first compute the orientation of a
    triangular cell:

       +                    +
       | \                  | \
       1   0               47   42
       |     \              |     \
       +--2---+             +--43--+
    FIAT canonical     Mapped example physical cell

    Suppose that the facets of the physical cell
    are canonically ordered as:

    C = [43, 42, 47]

    FIAT facet to Physical facet map is given by:

    M = [42, 47, 43]

    Then the orientation of the cell is computed as:

    C.index(M[0]) = 1; C.remove(M[0])
    C.index(M[1]) = 1; C.remove(M[1])
    C.index(M[2]) = 0; C.remove(M[2])

    o = (1 * 2!) + (1 * 1!) + (0 * 0!) = 3

    For npoints = 3, there are 6 DoFs:

        5                   0
        3 4                 1 3
        0 1 2               2 4 5
    FIAT canonical     Physical cell canonical

    The permutation associated with o = 3 then is:

    [2, 4, 5, 1, 3, 0]

    The output of this function contains one such permutation
    for each orientation for the given simplex dimension and
    the number of points along each axis.
    """
    from FIAT.polynomial_set import mis

    if npoints <= 0:
        return {o: [] for o in range(math.factorial(dim + 1))}
    a = np.array(sorted(mis(dim + 1, npoints - 1)), dtype=int)[:, ::-1]
    index_perms = sorted(itertools.permutations(range(dim + 1)))
    perms = {}
    for o, index_perm in enumerate(index_perms):
        perm = np.lexsort(np.transpose(a[:, index_perm]))
        perms[o] = perm.tolist()
    return perms


def _make_axis_perms_tensorproduct(cells, dim):
    """Return axis permutations for extrinsic orientations for the tensorproduct (sub)cell.

    :arg cells: The cells composing the tensorproduct cell.
    :arg dim: The (sub)dimensions of each component cell that makes up the tensorproduct (sub)cell.
    :returns: The tuple of axis permutations for all possible extrinsic orientations.

    Notes
    -----
    This is the single sources of extrinsic orientations and corresponding axis permutations.

    """
    from FIAT.reference_element import UFCInterval

    # Handle extrinsic orientations.
    # This is complex and we need to think to make this function more general.
    # One interesting case is pyramid x pyramid. There are two types of facets
    # in a pyramid cell, quad and triangle, and two types of intervals, ones
    # attached to quad (Iq) and ones attached to triangles (It). When we take
    # a tensor product of two pyramid cells, there are different kinds of tensor
    # product of intervals, i.e., Iq x Iq, Iq x It, It x Iq, It x It, and we
    # need a careful thought on how many possible extrinsic orientations we need
    # to consider for each.
    # For now we restrict ourselves to specific cases.
    nprod = len(cells)
    if len(set(cells)) == nprod:
        # All components have different cells.
        # Example: triangle x interval.
        #          dim == (2, 1) ->
        #          triangle x interval (1 possible extrinsic orientation).
        axis_perms = (tuple(range(nprod)), )  # Identity: no permutations
    elif len(set(cells)) == 1 and isinstance(cells[0], UFCInterval):
        # Tensor product of intervals.
        # Example: interval x interval x interval x interval
        #          dim == (0, 1, 1, 1) ->
        #          point x interval x interval x interval  (1! * 3! possible extrinsic orientations).
        axis_perms = sorted(itertools.permutations(range(nprod)))
        for idim, d in enumerate(dim):
            if d == 0:
                # idim-th component does not contribute to the extrinsic orientation.
                axis_perms = [ap for ap in axis_perms if ap[idim] == idim]
    else:
        # More general tensor product cells.
        # Example: triangle x quad x triangle x triangle x interval x interval
        #          dim == (2, 2, 2, 2, 1, 1) ->
        #          triangle x quad x triangle x triangle x interval x interval (3! * 1! * 2! possible extrinsic orientations).
        raise NotImplementedError(f"Unable to compose permutations for {' x '.join([str(cell) for cell in cells])}")
    return axis_perms


def make_entity_permutations_tensorproduct(cells, dim, o_p_maps):
    """Make orientation-permutation map for an entity of a tensor product cell.

    :arg cells: List of cells composing the tensor product cell.
    :arg dim: List of (sub)dimensions of the component cells that makes the tensor product (sub)cell.
    :arg o_p_maps: List of orientation-permutation maps of the component (sub)cells.
    :returns: The orientation-permutation map of the tensor product (sub)cell.

    Example
    -------

    .. code-block:: python3

        from FIAT.reference_element import UFCInterval
        from FIAT.orientation_utils import make_entity_permutations_tensorproduct


        cells = [UFCInterval(), UFCInterval()]
        m = make_entity_permutations_tensorproduct(cells, [1, 0], [{0: [0, 1], 1: [1, 0]}, {0: [0]}])
        print(m)
        # prints:
        # {(0, 0, 0): [0, 1],
        #  (0, 1, 0): [1, 0]}
        m = make_entity_permutations_tensorproduct(cells, [1, 1], [{0: [0, 1], 1: [1, 0]}, {0: [0, 1], 1: [1, 0]}])
        print(m)
        # prints:
        # {(0, 0, 0): [0, 1, 2, 3],
        #  (0, 0, 1): [1, 0, 3, 2],
        #  (0, 1, 0): [2, 3, 0, 1],
        #  (0, 1, 1): [3, 2, 1, 0],
        #  (1, 0, 0): [0, 2, 1, 3],
        #  (1, 0, 1): [2, 0, 3, 1],
        #  (1, 1, 0): [1, 3, 0, 2],
        #  (1, 1, 1): [3, 1, 2, 0]}

    """
    nprod = len(o_p_maps)
    assert len(cells) == nprod
    assert len(dim) == nprod
    axis_perms = _make_axis_perms_tensorproduct(cells, dim)
    o_tuple_perm_map = {}
    for eo, ap in enumerate(axis_perms):
        for o_tuple in itertools.product(*[o_p_map.keys() for o_p_map in o_p_maps]):
            ps = [o_p_map[o] for o_p_map, o in zip(o_p_maps, o_tuple)]
            shape = [len(p) for p in ps]
            for idim in range(len(ap)):
                shape[ap[idim]] = len(ps[idim])
            size = np.prod(shape)
            if size == 0:
                o_tuple_perm_map[(eo, ) + o_tuple] = []
            else:
                a = np.arange(size).reshape(shape)
                # Tensorproduct elements on a tensorproduct cell of intervals:
                # When we map the reference element to the physical element, we fisrt apply
                # the extrinsic orientation and then the intrinsic orientation.
                # Thus, to make the a.reshape(-1) trick work in the below,
                # we apply the inverse operation on a; we first apply the inverse of the
                # intrinsic orientation and then the inverse of the extrinsic orienataion.
                for idim, p in enumerate(ps):
                    # Note that p inverse = p for interval elements.
                    # Do not use p inverse (just use p) for elements on simplices
                    # as p already does what we want by construction.
                    a = a.swapaxes(0, ap[idim])[p, :].swapaxes(0, ap[idim])
                apinv = list(range(nprod))
                for idim in range(len(ap)):
                    apinv[ap[idim]] = idim
                a = np.moveaxis(a, range(nprod), apinv)
                o_tuple_perm_map[(eo, ) + o_tuple] = a.reshape(-1).tolist()
    return o_tuple_perm_map


def check_permutation_even_or_odd(_l):
    """Check if the given permutation is even or odd relative to ``[0, 1, ...)``.

    :arg _l: The permutation.
    :returns: ``0`` if even and ``1`` if odd.

    """
    assert isinstance(_l, Sequence), "The permutation must be Sequence"
    l = list(_l).copy()
    count = 0
    for i in range(len(l)):
        if l[i] != i:
            ii = l.index(i)
            l[ii] = l[i]
            l[i] = i
            count += 1
    assert l == sorted(list(_l))
    return count % 2


def make_cell_orientation_reflection_map_simplex(dim):
    npoints = 2
    o_p_map = make_entity_permutations_simplex(dim, npoints)
    cell_orientation_reflection_map = {}
    for o, p in o_p_map.items():
        reflected = check_permutation_even_or_odd(p)
        cell_orientation_reflection_map[o] = reflected
    assert cell_orientation_reflection_map[0] == 0, "Orientation 0 is expected to have no reflection"
    return cell_orientation_reflection_map


def make_cell_orientation_reflection_map_tensorproduct(cells):
    dim = [cell.get_dimension() for cell in cells]
    axis_perms = _make_axis_perms_tensorproduct(cells, dim)
    cell_orientation_reflection_map = {}
    for eo, ap in enumerate(axis_perms):
        # If ap has odd permutation, the associated extrinsic orientation (eo) introduces reflection.
        reflected_eo = check_permutation_even_or_odd(ap)
        for o_tuple in itertools.product(*[cell.cell_orientation_reflection_map().keys() for cell in cells]):
            reflected_io_list = [cell.cell_orientation_reflection_map()[o] for cell, o in zip(cells, o_tuple)]
            reflected = (reflected_eo + sum(reflected_io_list)) % 2
            cell_orientation_reflection_map[(eo, ) + o_tuple] = reflected
    return cell_orientation_reflection_map



import collections
import functools
import numbers
from functools import cached_property  # noqa: F401
from typing import Any

import numpy as np


def groupby(iterable, key=None):
    """Groups objects by their keys.

    :arg iterable: an iterable
    :arg key: key function

    :returns: list of (group key, list of group members) pairs
    """
    if key is None:
        key = lambda x: x
    groups = collections.OrderedDict()
    for elem in iterable:
        groups.setdefault(key(elem), []).append(elem)
    return groups.items()


def make_proxy_class(name, cls):
    """Constructs a proxy class for a given class.

    :arg name: name of the new proxy class
    :arg cls: the wrapee class to create a proxy for
    """
    def __init__(self, wrapee):
        self._wrapee = wrapee

    def make_proxy_property(name):
        def getter(self):
            return getattr(self._wrapee, name)
        return property(getter)

    dct = {'__init__': __init__}
    for attr in dir(cls):
        if not attr.startswith('_'):
            dct[attr] = make_proxy_property(attr)
    return type(name, (), dct)


# Implementation of dynamically scoped variables in Python.
class UnsetVariableError(LookupError):
    pass


_unset = object()


class DynamicallyScoped(object):
    """A dynamically scoped variable."""

    def __init__(self, default_value=_unset):
        if default_value is _unset:
            self._head = None
        else:
            self._head = (default_value, None)

    def let(self, value):
        return _LetBlock(self, value)

    @property
    def value(self):
        if self._head is None:
            raise UnsetVariableError("Dynamically scoped variable not set.")
        result, tail = self._head
        return result


class _LetBlock(object):
    """Context manager representing a dynamic scope."""

    def __init__(self, variable, value):
        self.variable = variable
        self.value = value
        self.state = None

    def __enter__(self):
        assert self.state is None
        value = self.value
        tail = self.variable._head
        scope = (value, tail)
        self.variable._head = scope
        self.state = scope

    def __exit__(self, exc_type, exc_value, traceback):
        variable = self.variable
        assert self.state is variable._head
        value, variable._head = variable._head
        self.state = None


@functools.singledispatch
def safe_repr(obj: Any) -> str:
    """Return a 'safe' repr for an object, accounting for floating point error.

    Parameters
    ----------
    obj :
        The object to produce a repr for.

    Returns
    -------
    str :
        A repr for the object.

    """
    raise TypeError(f"Cannot provide a safe repr for {type(obj).__name__}")


@safe_repr.register(str)
def _(text: str) -> str:
    return text


@safe_repr.register(numbers.Integral)
def _(num: numbers.Integral) -> str:
    return repr(num)


@safe_repr.register(numbers.Real)
def _(num: numbers.Real) -> str:
    # set roundoff to close-to-but-not-exactly machine epsilon
    precision = np.finfo(num).precision - 2
    return "{:.{prec}}".format(num, prec=precision)


@safe_repr.register(np.ndarray)
def _(array: np.ndarray) -> str:
    return f"{type(array).__name__}([{', '.join(map(safe_repr, array))}])"


@safe_repr.register(list)
def _(list_: list) -> str:
    return f"[{', '.join(map(safe_repr, list_))}]"


@safe_repr.register(tuple)
def _(tuple_: tuple) -> str:
    return f"({', '.join(map(safe_repr, tuple_))})"



"""Implementation of the Hu-Zhang finite elements."""
import FIAT
from gem import ListTensor
from finat.fiat_elements import FiatElement
from finat.physically_mapped import Citations, identity, PhysicallyMappedElement
from finat.aw import _facet_transform, _evaluation_transform


class HuZhang(PhysicallyMappedElement, FiatElement):
    def __init__(self, cell, degree=3, variant=None):
        if Citations is not None:
            Citations().register("Hu2015")
        self.variant = variant
        super().__init__(FIAT.HuZhang(cell, degree, variant=variant))

    def basis_transformation(self, coordinate_mapping):
        ndofs = self.space_dimension()
        V = identity(ndofs)

        sd = self.cell.get_spatial_dimension()
        W = _evaluation_transform(self.cell, coordinate_mapping)

        # Put into the right rows and columns.
        V[0:3, 0:3] = V[3:6, 3:6] = V[6:9, 6:9] = W
        ncomp = W.shape[0]
        num_verts = sd+1
        cur = num_verts * ncomp

        Vsub = _facet_transform(self.cell, self.degree-2, coordinate_mapping)
        fdofs = Vsub.shape[0]
        V[cur:cur+fdofs, cur:cur+fdofs] = Vsub
        cur += fdofs

        # internal DOFs
        if self.variant == "point":
            while cur < ndofs:
                V[cur:cur+ncomp, cur:cur+ncomp] = W
                cur += ncomp

        # RESCALING FOR CONDITIONING
        h = coordinate_mapping.cell_size()
        for e in range(num_verts):
            V[:, ncomp*e:ncomp*(e+1)] *= 1/(h[e] * h[e])

        # Note: that the edge DOFs are scaled by edge lengths in FIAT implies
        # that they are already have the necessary rescaling to improve
        # conditioning.
        return ListTensor(V.T)
